==> org-lsst-ccs-subsystem-focal-plane-main-1.0.15-sources.jar: org/lsst/ccs/subsystem/focalplane/FocalPlaneSubsystem.java <==
package org.lsst.ccs.subsystem.focalplane;

import java.io.Serializable;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.lsst.ccs.ConfigurationService;
import org.lsst.ccs.StateChangeListener;
import org.lsst.ccs.Subsystem;
import org.lsst.ccs.bootstrap.BootstrapResourceUtils;
import org.lsst.ccs.bus.data.AgentCategory;
import org.lsst.ccs.bus.data.DataProviderInfo;
import org.lsst.ccs.bus.data.AgentInfo;
import org.lsst.ccs.bus.data.KeyValueData;
import org.lsst.ccs.commons.annotations.LookupField;
import org.lsst.ccs.daq.ims.DAQException;
import org.lsst.ccs.daq.ims.ImageMetaData;
import org.lsst.ccs.daq.ims.StoreSimulation;
import org.lsst.ccs.daq.utilities.FitsHeaderKeywordData;
import org.lsst.ccs.daq.utilities.FitsService;
import org.lsst.ccs.description.ComponentLookup;
import org.lsst.ccs.description.ComponentNode;
import org.lsst.ccs.drivers.reb.ClientFactory;
import org.lsst.ccs.drivers.reb.GlobalClient;
import org.lsst.ccs.drivers.reb.REBException;
import org.lsst.ccs.drivers.reb.sim.ClientFactorySimulation;
import org.lsst.ccs.framework.HasLifecycle;
import org.lsst.ccs.imagenaming.service.ImageNameService;
import org.lsst.ccs.messaging.AgentPresenceListener;
import org.lsst.ccs.services.AgentCommandDictionaryService;
import org.lsst.ccs.services.AgentPropertiesService;
import org.lsst.ccs.services.AgentStateService;
import org.lsst.ccs.services.DataProviderDictionaryService;
import org.lsst.ccs.services.HasDataProviderInfos;
import org.lsst.ccs.services.alert.AlertService;
import org.lsst.ccs.subsystem.common.actions.RebPowerAction;
import org.lsst.ccs.subsystem.common.focalplane.data.HasFocalPlaneData;
import org.lsst.ccs.subsystem.focalplane.data.FocalPlaneDataGroup;
import org.lsst.ccs.subsystem.focalplane.data.ImageMetaDataEvent;
import org.lsst.ccs.subsystem.focalplane.states.FocalPlaneState;
import org.lsst.ccs.subsystem.rafts.REBDevice;
import org.lsst.ccs.subsystem.rafts.data.RaftException;
import org.lsst.ccs.subsystem.rafts.states.RebDeviceState;
import org.lsst.ccs.subsystem.rafts.states.RebValidationState;
import org.lsst.ccs.utilities.ccd.CCD;
import org.lsst.ccs.utilities.ccd.FocalPlane;
import org.lsst.ccs.utilities.ccd.Raft;
import org.lsst.ccs.utilities.ccd.Reb;
import org.lsst.ccs.utilities.location.Location;
import org.lsst.ccs.utilities.location.LocationSet;
import org.lsst.ccs.utilities.taitime.CCSTimeStamp;

public class FocalPlaneSubsystem extends Subsystem implements HasLifecycle, HasDataProviderInfos, ImageCoordinatorUtilities {

    private static final Logger LOG = Logger.getLogger(FocalPlaneSubsystem.class.getName());

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private final Map<String, REBDevice> rebDevices = new LinkedHashMap<>();

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    DataProviderDictionaryService dictionaryService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private SequencerConfig sequencerConfig;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private ImageNameService imageNameService;
       
    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AgentStateService agentStateService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AlertService agentAlertService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private ConfigurationService configService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AgentPropertiesService agentPropertiesService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    DataProviderDictionaryService dataProviderDictionaryService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AgentCommandDictionaryService agentCommandDictionaryService;
    
    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private ImageCoordinatorService imageCoordinatorService;    

    private Sequencers sequencers;
    private final FocalPlane focalPlaneGeometry;
    private ImageMessageHandling imageMessageHandling;

    private ComponentLookup lookupService;
    private final boolean isSimulationMode;

    public FocalPlaneSubsystem(FocalPlane geometry) {
        super("fp", AgentInfo.AgentType.WORKER);
        this.focalPlaneGeometry = geometry;
        getAgentInfo().getAgentProperties().setProperty("org.lsst.ccs.use.full.paths", "true");

        Properties props = BootstrapResourceUtils.getBootstrapSystemProperties();
        isSimulationMode = "simulation".equalsIgnoreCase(props.getProperty("org.lsst.ccs.run.mode", "normal"));
        if (isSimulationMode) {
            LOG.log(Level.INFO, "Running in simulation mode");
        }
    }

    @Override
    public void build() {
        agentPropertiesService.setAgentProperty(HasFocalPlaneData.AGENT_PROPERTY, HasFocalPlaneData.generatePropertyValue(FocalPlaneDataGroup.class));
        agentPropertiesService.setAgentProperty(AgentCategory.AGENT_CATEGORY_PROPERTY, AgentCategory.FOCAL_PLANE.name());
        agentPropertiesService.setAgentProperty(RebPowerAction.class.getCanonicalName(), "uses");

        // FIXME: What does this do?
        lookupService = this.getComponentLookup();
        ComponentNode thisComponent = lookupService.getComponentNodeForObject(this);

        // This is just avoiding modifying groovy?
        imageMessageHandling = new ImageMessageHandling();
        ComponentNode imageMessageHandlingNode = new ComponentNode("imageMessageHandling", imageMessageHandling);
        lookupService.addComponentNodeToLookup(thisComponent, imageMessageHandlingNode);

        for (Raft r : focalPlaneGeometry.getChildrenList()) {

            ComponentNode raftNode = new ComponentNode(r.getName(), new HardwareIdConfiguration());
            lookupService.addComponentNodeToLookup(thisComponent, raftNode);
            for (Reb reb : r.getRebs()) {
                String rebName = r.getName() + "/" + reb.getName();

                ComponentNode rebHardwareNode = new ComponentNode(reb.getName() + "_hardware", new HardwareIdConfiguration());
                lookupService.addComponentNodeToLookup(raftNode, rebHardwareNode);

                ComponentNode rebComponent = lookupService.getNodeByPath(rebName);
                for (CCD ccd : reb.getCCDs()) {
                    String sensorName = ccd.getName();
                    ComponentNode sensorNode = new ComponentNode(sensorName, new CCDHardwareIdConfiguration());
                    lookupService.addComponentNodeToLookup(rebComponent, sensorNode);
                }
            }
        }
        //Register the FocalPlaneState
        agentStateService.registerState(FocalPlaneState.class, "The state of the Focal Plane", this);

    }

    @Override
    public void init() {
        // This used to be done by global proc, now we need to do it here.
        ClientFactory clientFactory = isSimulationMode ? new ClientFactorySimulation(true) : new ClientFactory();
        for (REBDevice rebDevice : rebDevices.values()) {
            rebDevice.setClientFactory(clientFactory);
        }
        if (isSimulationMode) {
            GlobalClient gbl = new GlobalClient();
            gbl.setClientFactory(clientFactory);
            try {
                gbl.open(sequencerConfig.getPartition());
                StoreSimulation.instance().addTriggerListener((int opcode, ImageMetaData meta, Map<Location.LocationType, int[]> registerLists) -> {
                    try {
                        if (meta == null) {
                            gbl.startSequencer(opcode);
                        } else {
                            for (Map.Entry<Location.LocationType, int[]> entry : registerLists.entrySet()) {
                                gbl.setRegisterlist(entry.getKey().getCCDCount(), entry.getValue());
                            }
                            int[] ids = new LocationSet(meta.getLocations()).asIntArray();
                            gbl.acquireImage(meta.getName(), meta.getCreationFolderName(), opcode, meta.getAnnotation(), ids);
                        }
                    } catch (REBException x) {
                        LOG.log(Level.SEVERE, "Error simulating trigger", x);
                    }
                });
            } catch (REBException x) {
                throw new RuntimeException("Failed to initialize simulation", x);
            }
        }
    }

    @Override
    public void postInit() {

        sequencers = new Sequencers(this, sequencerConfig);
        agentStateService.updateAgentState(FocalPlaneState.NEEDS_CLEAR);
        dataProviderDictionaryService.registerClass(ImageMetaDataEvent.class, ImageMetaDataEvent.EVENT_KEY);

        rebDevices.forEach((rebName, device) -> {
            Reb reb = focalPlaneGeometry.getReb(rebName);
            device.setRebGeometry(reb);
            sequencers.add(reb, device.getSequencer());
        });

        agentStateService.addStateChangeListener(new StateChangeListener() {
            @Override
            public void stateChanged(CCSTimeStamp stateTransitionTimestamp, Object device, Enum<?> newState, Enum<?> oldState) {
                if (newState == RebValidationState.VALID) {
                    final REBDevice rebDevice = (REBDevice) device;
                    String path = lookupService.getNameOfComponent(device);
                    try {
                        FirmwareVersion version = new FirmwareVersion(rebDevice.getHwVersion());
                        if (version.hasLookAtMeBug()) {
                            LOG.log(Level.INFO, "Disabled LAM for firmware {0}", version);
                            // FIXME: Temporary workaround for LSSTCCSRAFTS-529
                            try {
                                rebDevice.setRegister(0x17, new int[]{0});
                            } catch (RaftException x) {
                                LOG.log(Level.SEVERE, "Could not disable LAM for " + rebDevice.getName(), x);
                            }
                        }

                        sequencers.load(focalPlaneGeometry.getReb(path), version);

                        Object result = rebDevice.loadAspics(true);
                        LOG.log(Level.INFO, "Loaded aspics for {0} result {1}", new Object[]{rebDevice.getName(), result});

                    } catch (Exception x) {
                        LOG.log(Level.WARNING, "Exception while loading ASPICS", x);
                    }
                }
            }
        }, RebDeviceState.class, RebValidationState.class);

        agentCommandDictionaryService.addCommandSetToObject(new LSE71Commands(this), "");
        agentCommandDictionaryService.addCommandSetToObject(new ScriptingCommands(this), "");
        agentCommandDictionaryService.addCommandSetToObject(new FocalPlaneCommands(this), "");

        //This is needed to publish the Alert summary when a Power Subsystem 
        //joins the buses.
        getMessagingAccess().getAgentPresenceManager().addAgentPresenceListener(
                new AgentPresenceListener() {
            @Override
            public void connected(AgentInfo... agents) {
                for (AgentInfo ai : agents) {
                    if (ai.getAgentProperty(AgentCategory.AGENT_CATEGORY_PROPERTY, "").equals(AgentCategory.POWER.name())) {
                        LOG.log(Level.INFO, "Publishing raised alert summary for {0}", new Object[]{ai.getName()});
                        getScheduler().execute(() -> {
                            agentAlertService.publishRaisedAlertSummary();
                        });
                        break;
                    }
                }
            }
        });
    }

    @Override
    public void finalizeDictionary() {
        //Manipulate the dictionary here: 
        for (DataProviderInfo data : dictionaryService.getDataProviderDictionary().getDataProviderInfos()) {
            FocalPlaneDataGroup dataGroup = FocalPlaneDataGroup.findFocalPlaneDataGroup(data);
            if (dataGroup != null) {
                dataGroup.addAttributesToDataInfo(data);
            }
        }
    }

    @Override
    public void postStart() {
        configService.addConfigurationListener(new ConfigListener(this));
        loadSequencers();
    }

    void loadSequencers() {
        try {
            sequencers.load();
        } catch (RaftException | REBException | DAQException x) {
            //TODO: Go into fault state?
            throw new RuntimeException("Error configuring sequencers", x);
        }
    }

    void loadASPICS() {
        for (REBDevice device : rebDevices.values()) {
            if (device.isOnline() && device.isSerialNumValid()) {
                try {
                    Object result = device.loadAspics(true);
                    LOG.log(Level.INFO, "Loaded aspics for {0} result {1}", new Object[]{device.getFullName(), result});
                } catch (Exception ex) {
                    //TODO: Go into fault state?
                    LOG.log(Level.SEVERE, "Error while loading ASPICS", ex);
                }
            }
        }
    }

    boolean isRebOnlineAndValid(Reb reb) {
        return stateService.isComponentInState(reb.getFullName(), RebValidationState.VALID);
    }

    FirmwareVersion getFirmwareVersionForReb(Reb reb) {
        return new FirmwareVersion(rebDevices.get(reb.getFullName()).getHwVersion());
    }

    public Sequencers getSequencers() {
        return sequencers;
    }

    ImageNameService getImageNameService() {
        return imageNameService;
    }

    Map<String, REBDevice> getRebDevices() {
        return rebDevices;
    }

    void loadSequencerParameters() {
        sequencers.loadSequencerParameters();
    }

    @Override
    public void sendEvent(String key, Serializable event) {
        KeyValueData kvd = new KeyValueData(key, event);
        publishSubsystemDataOnStatusBus(kvd);
        LOG.log(Level.INFO, "Sent: {0}", event);
    }


    void handleMetaData(Map<String, Serializable> headersMap) {
        // TODO: Fix me
    }

    @Override
    public void setStateIf(FocalPlaneState currentState, FocalPlaneState newState) {
        // FIXME: This should be made atomic
        if (agentStateService.getState(FocalPlaneState.class) == currentState) {
           agentStateService.updateAgentState(newState);
        }
    }

    ImageCoordinatorService getImageCoordinatorService() {
        return imageCoordinatorService;
    }

    void setState(FocalPlaneState newState) {
        agentStateService.updateAgentState(newState);
    }

    /**
     * Method called to add meta-data to images. The meta-data can got both into the image-handlers, and
     * into the image image database.
     * @param headersMap 
     */
    void setHeaderKeywords(Map<String, Serializable> headersMap) {
        FitsHeaderKeywordData data = new FitsHeaderKeywordData();
        for ( Map.Entry<String,Serializable> e : headersMap.entrySet() ) {
            data.addHeaderKeywordValue(FitsService.COMMON_HEADER_NAME, e.getKey(), e.getValue(), false);
        }
        publishSubsystemDataOnStatusBus(data.getKeyValueData());
        imageCoordinatorService.addMetaData(headersMap);
    }
}

==> org-lsst-ccs-subsystem-focal-plane-main-1.0.16-sources.jar: org/lsst/ccs/subsystem/focalplane/FocalPlaneSubsystem.java <==
package org.lsst.ccs.subsystem.focalplane;

import java.io.Serializable;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.lsst.ccs.ConfigurationService;
import org.lsst.ccs.StateChangeListener;
import org.lsst.ccs.Subsystem;
import org.lsst.ccs.bootstrap.BootstrapResourceUtils;
import org.lsst.ccs.bus.data.AgentCategory;
import org.lsst.ccs.bus.data.DataProviderInfo;
import org.lsst.ccs.bus.data.AgentInfo;
import org.lsst.ccs.bus.data.KeyValueData;
import org.lsst.ccs.commons.annotations.LookupField;
import org.lsst.ccs.daq.ims.DAQException;
import org.lsst.ccs.daq.ims.ImageMetaData;
import org.lsst.ccs.daq.ims.StoreSimulation;
import org.lsst.ccs.daq.utilities.FitsHeaderKeywordData;
import org.lsst.ccs.daq.utilities.FitsService;
import org.lsst.ccs.description.ComponentLookup;
import org.lsst.ccs.description.ComponentNode;
import org.lsst.ccs.drivers.reb.ClientFactory;
import org.lsst.ccs.drivers.reb.GlobalClient;
import org.lsst.ccs.drivers.reb.REBException;
import org.lsst.ccs.drivers.reb.sim.ClientFactorySimulation;
import org.lsst.ccs.framework.HasLifecycle;
import org.lsst.ccs.imagenaming.service.ImageNameService;
import org.lsst.ccs.messaging.AgentPresenceListener;
import org.lsst.ccs.services.AgentCommandDictionaryService;
import org.lsst.ccs.services.AgentPropertiesService;
import org.lsst.ccs.services.AgentStateService;
import org.lsst.ccs.services.DataProviderDictionaryService;
import org.lsst.ccs.services.HasDataProviderInfos;
import org.lsst.ccs.services.alert.AlertService;
import org.lsst.ccs.subsystem.common.actions.RebPowerAction;
import org.lsst.ccs.subsystem.common.focalplane.data.HasFocalPlaneData;
import org.lsst.ccs.subsystem.focalplane.data.FocalPlaneDataGroup;
import org.lsst.ccs.subsystem.focalplane.data.ImageMetaDataEvent;
import org.lsst.ccs.subsystem.focalplane.states.FocalPlaneState;
import org.lsst.ccs.subsystem.rafts.REBDevice;
import org.lsst.ccs.subsystem.rafts.data.RaftException;
import org.lsst.ccs.subsystem.rafts.states.RebDeviceState;
import org.lsst.ccs.subsystem.rafts.states.RebValidationState;
import org.lsst.ccs.utilities.ccd.CCD;
import org.lsst.ccs.utilities.ccd.FocalPlane;
import org.lsst.ccs.utilities.ccd.Raft;
import org.lsst.ccs.utilities.ccd.Reb;
import org.lsst.ccs.utilities.location.Location;
import org.lsst.ccs.utilities.location.LocationSet;
import org.lsst.ccs.utilities.taitime.CCSTimeStamp;

public class FocalPlaneSubsystem extends Subsystem implements HasLifecycle, HasDataProviderInfos {

    private static final Logger LOG = Logger.getLogger(FocalPlaneSubsystem.class.getName());

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private final Map<String, REBDevice> rebDevices = new LinkedHashMap<>();

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    DataProviderDictionaryService dictionaryService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private SequencerConfig sequencerConfig;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private ImageNameService imageNameService;
       
    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AgentStateService agentStateService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AlertService agentAlertService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private ConfigurationService configService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AgentPropertiesService agentPropertiesService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    DataProviderDictionaryService dataProviderDictionaryService;

    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private AgentCommandDictionaryService agentCommandDictionaryService;
    
    @LookupField(strategy = LookupField.Strategy.DESCENDANTS)
    private ImageCoordinatorService imageCoordinatorService;    

    private Sequencers sequencers;
    private final FocalPlane focalPlaneGeometry;
    private ImageMessageHandling imageMessageHandling;

    private ComponentLookup lookupService;
    private final boolean isSimulationMode;

    public FocalPlaneSubsystem(FocalPlane geometry) {
        super("fp", AgentInfo.AgentType.WORKER);
        this.focalPlaneGeometry = geometry;
        getAgentInfo().getAgentProperties().setProperty("org.lsst.ccs.use.full.paths", "true");

        Properties props = BootstrapResourceUtils.getBootstrapSystemProperties();
        isSimulationMode = "simulation".equalsIgnoreCase(props.getProperty("org.lsst.ccs.run.mode", "normal"));
        if (isSimulationMode) {
            LOG.log(Level.INFO, "Running in simulation mode");
        }
    }

    @Override
    public void build() {
        agentPropertiesService.setAgentProperty(HasFocalPlaneData.AGENT_PROPERTY, HasFocalPlaneData.generatePropertyValue(FocalPlaneDataGroup.class));
        agentPropertiesService.setAgentProperty(AgentCategory.AGENT_CATEGORY_PROPERTY, AgentCategory.FOCAL_PLANE.name());
        agentPropertiesService.setAgentProperty(RebPowerAction.class.getCanonicalName(), "uses");

        // FIXME: What does this do?
        lookupService = this.getComponentLookup();
        ComponentNode thisComponent = lookupService.getComponentNodeForObject(this);

        // This is just avoiding modifying groovy?
        imageMessageHandling = new ImageMessageHandling();
        ComponentNode imageMessageHandlingNode = new ComponentNode("imageMessageHandling", imageMessageHandling);
        lookupService.addComponentNodeToLookup(thisComponent, imageMessageHandlingNode);

        for (Raft r : focalPlaneGeometry.getChildrenList()) {

            ComponentNode raftNode = new ComponentNode(r.getName(), new HardwareIdConfiguration());
            lookupService.addComponentNodeToLookup(thisComponent, raftNode);
            for (Reb reb : r.getRebs()) {
                String rebName = r.getName() + "/" + reb.getName();

                ComponentNode rebHardwareNode = new ComponentNode(reb.getName() + "_hardware", new HardwareIdConfiguration());
                lookupService.addComponentNodeToLookup(raftNode, rebHardwareNode);

                ComponentNode rebComponent = lookupService.getNodeByPath(rebName);
                for (CCD ccd : reb.getCCDs()) {
                    String sensorName = ccd.getName();
                    ComponentNode sensorNode = new ComponentNode(sensorName, new CCDHardwareIdConfiguration());
                    lookupService.addComponentNodeToLookup(rebComponent, sensorNode);
                }
            }
        }
        //Register the FocalPlaneState
        agentStateService.registerState(FocalPlaneState.class, "The state of the Focal Plane", this);

    }

    @Override
    public void init() {
        // This used to be done by global proc, now we need to do it here.
        ClientFactory clientFactory = isSimulationMode ? new ClientFactorySimulation(true) : new ClientFactory();
        for (REBDevice rebDevice : rebDevices.values()) {
            rebDevice.setClientFactory(clientFactory);
        }
        if (isSimulationMode) {
            GlobalClient gbl = new GlobalClient();
            gbl.setClientFactory(clientFactory);
            try {
                gbl.open(sequencerConfig.getPartition());
                StoreSimulation.instance().addTriggerListener((int opcode, ImageMetaData meta, Map<Location.LocationType, int[]> registerLists) -> {
                    try {
                        if (meta == null) {
                            gbl.startSequencer(opcode);
                        } else {
                            for (Map.Entry<Location.LocationType, int[]> entry : registerLists.entrySet()) {
                                gbl.setRegisterlist(entry.getKey().getCCDCount(), entry.getValue());
                            }
                            int[] ids = new LocationSet(meta.getLocations()).asIntArray();
                            gbl.acquireImage(meta.getName(), meta.getCreationFolderName(), opcode, meta.getAnnotation(), ids);
                        }
                    } catch (REBException x) {
                        LOG.log(Level.SEVERE, "Error simulating trigger", x);
                    }
                });
            } catch (REBException x) {
                throw new RuntimeException("Failed to initialize simulation", x);
            }
        }
    }

    @Override
    public void postInit() {

        sequencers = new Sequencers(this, sequencerConfig);
        agentStateService.updateAgentState(FocalPlaneState.NEEDS_CLEAR);
        dataProviderDictionaryService.registerClass(ImageMetaDataEvent.class, ImageMetaDataEvent.EVENT_KEY);

        rebDevices.forEach((rebName, device) -> {
            Reb reb = focalPlaneGeometry.getReb(rebName);
            device.setRebGeometry(reb);
            sequencers.add(reb, device.getSequencer());
        });

        agentStateService.addStateChangeListener(new StateChangeListener() {
            @Override
            public void stateChanged(CCSTimeStamp stateTransitionTimestamp, Object device, Enum<?> newState, Enum<?> oldState) {
                if (newState == RebValidationState.VALID) {
                    final REBDevice rebDevice = (REBDevice) device;
                    String path = lookupService.getNameOfComponent(device);
                    try {
                        FirmwareVersion version = new FirmwareVersion(rebDevice.getHwVersion());
                        if (version.hasLookAtMeBug()) {
                            LOG.log(Level.INFO, "Disabled LAM for firmware {0}", version);
                            // FIXME: Temporary workaround for LSSTCCSRAFTS-529
                            try {
                                rebDevice.setRegister(0x17, new int[]{0});
                            } catch (RaftException x) {
                                LOG.log(Level.SEVERE, "Could not disable LAM for " + rebDevice.getName(), x);
                            }
                        }

                        sequencers.load(focalPlaneGeometry.getReb(path), version);

                        Object result = rebDevice.loadAspics(true);
                        LOG.log(Level.INFO, "Loaded aspics for {0} result {1}", new Object[]{rebDevice.getName(), result});

                    } catch (Exception x) {
                        LOG.log(Level.WARNING, "Exception while loading ASPICS", x);
                    }
                }
            }
        }, RebDeviceState.class, RebValidationState.class);

        agentCommandDictionaryService.addCommandSetToObject(new LSE71Commands(this), "");
        agentCommandDictionaryService.addCommandSetToObject(new ScriptingCommands(this), "");
        agentCommandDictionaryService.addCommandSetToObject(new FocalPlaneCommands(this), "");

        //This is needed to publish the Alert summary when a Power Subsystem 
        //joins the buses.
        getMessagingAccess().getAgentPresenceManager().addAgentPresenceListener(
                new AgentPresenceListener() {
            @Override
            public void connected(AgentInfo... agents) {
                for (AgentInfo ai : agents) {
                    if (ai.getAgentProperty(AgentCategory.AGENT_CATEGORY_PROPERTY, "").equals(AgentCategory.POWER.name())) {
                        LOG.log(Level.INFO, "Publishing raised alert summary for {0}", new Object[]{ai.getName()});
                        getScheduler().execute(() -> {
                            agentAlertService.publishRaisedAlertSummary();
                        });
                        break;
                    }
                }
            }
        });
    }

    @Override
    public void finalizeDictionary() {
        //Manipulate the dictionary here: 
        for (DataProviderInfo data : dictionaryService.getDataProviderDictionary().getDataProviderInfos()) {
            FocalPlaneDataGroup dataGroup = FocalPlaneDataGroup.findFocalPlaneDataGroup(data);
            if (dataGroup != null) {
                dataGroup.addAttributesToDataInfo(data);
            }
        }
    }

    @Override
    public void postStart() {
        configService.addConfigurationListener(new ConfigListener(this));
        loadSequencers();
    }

    void loadSequencers() {
        try {
            sequencers.load();
        } catch (RaftException | REBException | DAQException x) {
            //TODO: Go into fault state?
            throw new RuntimeException("Error configuring sequencers", x);
        }
    }

    void loadASPICS() {
        for (REBDevice device : rebDevices.values()) {
            if (device.isOnline() && device.isSerialNumValid()) {
                try {
                    Object result = device.loadAspics(true);
                    LOG.log(Level.INFO, "Loaded aspics for {0} result {1}", new Object[]{device.getFullName(), result});
                } catch (Exception ex) {
                    //TODO: Go into fault state?
                    LOG.log(Level.SEVERE, "Error while loading ASPICS", ex);
                }
            }
        }
    }

    boolean isRebOnlineAndValid(Reb reb) {
        return stateService.isComponentInState(reb.getFullName(), RebValidationState.VALID);
    }

    FirmwareVersion getFirmwareVersionForReb(Reb reb) {
        return new FirmwareVersion(rebDevices.get(reb.getFullName()).getHwVersion());
    }

    public Sequencers getSequencers() {
        return sequencers;
    }

    ImageNameService getImageNameService() {
        return imageNameService;
    }

    Map<String, REBDevice> getRebDevices() {
        return rebDevices;
    }

    void loadSequencerParameters() {
        sequencers.loadSequencerParameters();
    }

    void sendEvent(String key, Serializable event) {
        KeyValueData kvd = new KeyValueData(key, event);
        publishSubsystemDataOnStatusBus(kvd);
        LOG.log(Level.INFO, "Sent: {0}", event);
    }

    void setState(FocalPlaneState newState) {
        agentStateService.updateAgentState(newState);
    }

    void setStateIf(FocalPlaneState currentState, FocalPlaneState newState) {
        synchronized (agentStateService.getStateLock()) {
            if (agentStateService.getState(FocalPlaneState.class) == currentState) {
               agentStateService.updateAgentState(newState);
            }
        }
    }

    ImageCoordinatorService getImageCoordinatorService() {
        return imageCoordinatorService;
    }

    /**
     * Method called to add meta-data to images. The meta-data can got both into the image-handlers, and
     * into the image image database.
     * @param headersMap 
     */
    void setHeaderKeywords(Map<String, Serializable> headersMap) {
        FitsHeaderKeywordData data = new FitsHeaderKeywordData();
        for ( Map.Entry<String,Serializable> e : headersMap.entrySet() ) {
            data.addHeaderKeywordValue(FitsService.COMMON_HEADER_NAME, e.getKey(), e.getValue(), false);
        }
        publishSubsystemDataOnStatusBus(data.getKeyValueData());
        imageCoordinatorService.addMetaData(headersMap);
    }
    
    /**
     * Ugly way to put the ASPICS into transparent mode. Required for LSSTCCSRAFTS-567.
     * @param reb
     * @param tranparentMode 
     */
    void setTransparentMode(Reb reb, boolean tranparentMode) {
        rebDevices.get(reb.getFullName()).setAllAspicTM(tranparentMode);
    }
}
